@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2020, the neonavigation authors
+ * Copyright (c) 2014-2023, the neonavigation authors
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -48,6 +48,8 @@
 #include <nav_msgs/GetPlan.h>
 #include <nav_msgs/OccupancyGrid.h>
 #include <nav_msgs/Path.h>
+#include <neonavigation_metrics_msgs/Metrics.h>
+#include <neonavigation_metrics_msgs/helper.h>
 #include <planner_cspace_msgs/PlannerStatus.h>
 #include <sensor_msgs/PointCloud.h>
 #include <std_srvs/Empty.h>
@@ -74,6 +76,7 @@
 #include <planner_cspace/planner_3d/motion_cache.h>
 #include <planner_cspace/planner_3d/path_interpolator.h>
 #include <planner_cspace/planner_3d/rotation_cache.h>
+#include <planner_cspace/planner_3d/distance_map.h>
 
 namespace planner_cspace
 {
@@ -102,6 +105,7 @@ protected:
   ros::Publisher pub_start_;
   ros::Publisher pub_end_;
   ros::Publisher pub_status_;
+  ros::Publisher pub_metrics_;
   ros::ServiceServer srs_forget_;
   ros::ServiceServer srs_make_plan_;
 
@@ -118,22 +122,22 @@ protected:
   Astar::Gridmap<char, 0x80> cm_rough_base_;
   Astar::Gridmap<char, 0x80> cm_hyst_;
   Astar::Gridmap<char, 0x80> cm_updates_;
-  Astar::Gridmap<float> cost_estim_cache_;
   CostmapBBF bbf_costmap_;
+  DistanceMap cost_estim_cache_;
 
   GridAstarModel3D::Ptr model_;
-  std::array<float, 1024> euclid_cost_lin_cache_;
 
   costmap_cspace_msgs::MapMetaData3D map_info_;
+  costmap_cspace_msgs::CSpace3DUpdate::ConstPtr map_update_retained_;
   std_msgs::Header map_header_;
   float freq_;
   float freq_min_;
+  float search_timeout_abort_;
   float search_range_;
   bool antialias_start_;
   int range_;
   int local_range_;
   double local_range_f_;
-  int longcut_range_;
   double longcut_range_f_;
   int esc_range_;
   int esc_angle_;
@@ -148,6 +152,7 @@ protected:
   bool has_goal_;
   bool has_start_;
   bool has_hysteresis_map_;
+  std::vector<Astar::Vec> hyst_updated_cells_;
   bool goal_updated_;
   bool remember_updates_;
   bool fast_map_update_;
@@ -169,7 +174,6 @@ protected:
   int max_retry_num_;
 
   int num_task_;
-  int num_cost_estim_task_;
 
   // Cost weights
   CostCoeff cc_;
@@ -185,13 +189,13 @@ protected:
   int goal_tolerance_ang_;
 
   planner_cspace_msgs::PlannerStatus status_;
+  neonavigation_metrics_msgs::Metrics metrics_;
 
   bool find_best_;
   float sw_wait_;
   geometry_msgs::PoseStamped sw_pos_;
   bool is_path_switchback_;
 
-  float rough_cost_max_;
   bool rough_;
 
   bool force_goal_orientation_;
@@ -218,6 +222,48 @@ protected:
 
     return true;
   }
+  enum class DiscretePoseStatus
+  {
+    OK,
+    RELOCATED,
+    IN_ROCK,
+    OUT_OF_MAP,
+  };
+  template <class T>
+  DiscretePoseStatus relocateDiscretePoseIfNeededImpl(const T& cm,
+                                                      const int tolerance_range,
+                                                      const int tolerance_angle,
+                                                      Astar::Vec& pose_discrete) const
+  {
+    if (!cm.validate(pose_discrete, range_))
+    {
+      return DiscretePoseStatus::OUT_OF_MAP;
+    }
+    if (cm[pose_discrete] == 100)
+    {
+      if (searchAvailablePos(cm, pose_discrete, tolerance_range, tolerance_angle))
+        return DiscretePoseStatus::RELOCATED;
+      else
+        return DiscretePoseStatus::IN_ROCK;
+    }
+    return DiscretePoseStatus::OK;
+  }
+  DiscretePoseStatus relocateDiscretePoseIfNeeded(Astar::Vec& pose_discrete,
+                                                  const int tolerance_range,
+                                                  const int tolerance_angle,
+                                                  bool use_cm_rough = false) const
+  {
+    if (use_cm_rough)
+    {
+      pose_discrete[2] = 0;
+      return relocateDiscretePoseIfNeededImpl(cm_rough_, tolerance_range, tolerance_angle, pose_discrete);
+    }
+    else
+    {
+      return relocateDiscretePoseIfNeededImpl(cm_, tolerance_range, tolerance_angle, pose_discrete);
+    }
+  }
+
   bool cbMakePlan(nav_msgs::GetPlan::Request& req,
                   nav_msgs::GetPlan::Response& res)
   {
@@ -241,26 +287,44 @@ protected:
     grid_metric_converter::metric2Grid(
         map_info_, s[0], s[1], s[2],
         req.start.pose.position.x, req.start.pose.position.y, tf2::getYaw(req.start.pose.orientation));
-    s[2] = 0;
     grid_metric_converter::metric2Grid(
         map_info_, e[0], e[1], e[2],
         req.goal.pose.position.x, req.goal.pose.position.y, tf2::getYaw(req.goal.pose.orientation));
-    e[2] = 0;
+    ROS_INFO("Path plan from (%d, %d) to (%d, %d)", s[0], s[1], e[0], e[1]);
 
-    if (!(cm_rough_.validate(s, range_) && cm_rough_.validate(e, range_)))
+    const int tolerance_range = std::lround(req.tolerance / map_info_.linear_resolution);
+    const DiscretePoseStatus start_status = relocateDiscretePoseIfNeeded(s, tolerance_range, tolerance_angle_, true);
+    const DiscretePoseStatus goal_status = relocateDiscretePoseIfNeeded(e, tolerance_range, tolerance_angle_, true);
+    switch (start_status)
     {
-      ROS_ERROR("Given start or goal is not on the map.");
-      return false;
+      case DiscretePoseStatus::OUT_OF_MAP:
+        ROS_ERROR("Given start is not on the map.");
+        return false;
+      case DiscretePoseStatus::IN_ROCK:
+        ROS_ERROR("Given start is in Rock.");
+        return false;
+      case DiscretePoseStatus::RELOCATED:
+        ROS_INFO("Given start is moved (%d, %d)", s[0], s[1]);
+        break;
+      default:
+        break;
     }
-    else if (cm_rough_[s] == 100 || cm_rough_[e] == 100)
+    switch (goal_status)
     {
-      ROS_ERROR(
-          "Given start or goal is in Rock. (start: %d, end: %d)",
-          cm_rough_[s], cm_rough_[e]);
-      return false;
+      case DiscretePoseStatus::OUT_OF_MAP:
+        ROS_ERROR("Given goal is not on the map.");
+        return false;
+      case DiscretePoseStatus::IN_ROCK:
+        ROS_ERROR("Given goal is in Rock.");
+        return false;
+      case DiscretePoseStatus::RELOCATED:
+        ROS_INFO("Given goal is moved (%d, %d)", e[0], e[1]);
+        break;
+      default:
+        break;
     }
 
-    const auto cb_progress = [](const std::list<Astar::Vec>& path_grid)
+    const auto cb_progress = [](const std::list<Astar::Vec>& /* path_grid */, const SearchStats& /* stats */)
     {
       return true;
     };
@@ -351,6 +415,7 @@ protected:
         return false;
       }
       status_.status = planner_cspace_msgs::PlannerStatus::DOING;
+      status_.header.stamp = ros::Time::now();
       pub_status_.publish(status_);
       diag_updater_.update();
     }
@@ -364,169 +429,10 @@ protected:
     }
     return true;
   }
-  void fillCostmap(
-      reservable_priority_queue<Astar::PriorityVec>& open,
-      Astar::Gridmap<float>& g,
-      const Astar::Vec& s, const Astar::Vec& e)
-  {
-    const Astar::Vec s_rough(s[0], s[1], 0);
 
-    struct SearchDiffs
-    {
-      Astar::Vec d;
-      std::vector<Astar::Vec> pos;
-      float grid_to_len;
-    };
-    std::vector<SearchDiffs> search_diffs;
-    {
-      Astar::Vec d;
-      d[2] = 0;
-      const int range_rough = 4;
-      for (d[0] = -range_rough; d[0] <= range_rough; d[0]++)
-      {
-        for (d[1] = -range_rough; d[1] <= range_rough; d[1]++)
-        {
-          if (d[0] == 0 && d[1] == 0)
-            continue;
-          if (d.sqlen() > range_rough * range_rough)
-            continue;
-
-          SearchDiffs diffs;
-
-          const float grid_to_len = d.gridToLenFactor();
-          const int dist = d.len();
-          const float dpx = static_cast<float>(d[0]) / dist;
-          const float dpy = static_cast<float>(d[1]) / dist;
-          Astar::Vecf pos(0, 0, 0);
-          for (int i = 0; i < dist; i++)
-          {
-            Astar::Vec ipos(pos);
-            if (diffs.pos.size() == 0 || diffs.pos.back() != ipos)
-            {
-              diffs.pos.push_back(std::move(ipos));
-            }
-            pos[0] += dpx;
-            pos[1] += dpy;
-          }
-          diffs.grid_to_len = grid_to_len;
-          diffs.d = d;
-          search_diffs.push_back(std::move(diffs));
-        }
-      }
-    }
-
-    std::vector<Astar::PriorityVec> centers;
-    centers.reserve(num_cost_estim_task_);
-
-#pragma omp parallel
-    {
-      std::vector<Astar::GridmapUpdate> updates;
-      updates.reserve(num_cost_estim_task_ * search_diffs.size() / omp_get_num_threads());
-
-      const float range_overshoot = ec_[0] * (range_ + local_range_ + longcut_range_);
-
-      while (true)
-      {
-#pragma omp barrier
-#pragma omp single
-        {
-          centers.clear();
-          for (size_t i = 0; i < static_cast<size_t>(num_cost_estim_task_);)
-          {
-            if (open.size() < 1)
-              break;
-            Astar::PriorityVec center(open.top());
-            open.pop();
-            if (center.p_raw_ > g[center.v_])
-              continue;
-            if (center.p_raw_ - range_overshoot > g[s_rough])
-              continue;
-            centers.emplace_back(std::move(center));
-            ++i;
-          }
-        }  // omp single
-
-        if (centers.size() == 0)
-          break;
-        updates.clear();
-
-#pragma omp for schedule(static)
-        for (auto it = centers.cbegin(); it < centers.cend(); ++it)
-        {
-          const Astar::Vec p = it->v_;
-
-          for (const SearchDiffs& ds : search_diffs)
-          {
-            const Astar::Vec d = ds.d;
-            const Astar::Vec next = p + d;
-
-            if (static_cast<size_t>(next[0]) >= static_cast<size_t>(map_info_.width) ||
-                static_cast<size_t>(next[1]) >= static_cast<size_t>(map_info_.height))
-              continue;
-
-            float cost = model_->euclidCostRough(d);
-
-            const float gnext = g[next];
-
-            if (gnext < g[p] + cost)
-            {
-              // Skip as this search task has no chance to find better way.
-              continue;
-            }
-
-            {
-              float sum = 0, sum_hist = 0;
-              bool collision = false;
-              for (const auto& d : ds.pos)
-              {
-                const Astar::Vec pos = p + d;
-                const char c = cm_rough_[pos];
-                if (c > 99)
-                {
-                  collision = true;
-                  break;
-                }
-                sum += c;
-                sum_hist += bbf_costmap_.getCost(pos);
-              }
-              if (collision)
-                continue;
-              cost +=
-                  (map_info_.linear_resolution * ds.grid_to_len / 100.0) *
-                  (sum * cc_.weight_costmap_ + sum_hist * cc_.weight_remembered_);
-
-              if (cost < 0)
-              {
-                cost = 0;
-                ROS_WARN_THROTTLE(1.0, "Negative cost value is detected. Limited to zero.");
-              }
-            }
-
-            const float cost_next = it->p_raw_ + cost;
-            if (gnext > cost_next)
-            {
-              updates.emplace_back(p, next, cost_next, cost_next);
-            }
-          }
-        }
-#pragma omp barrier
-#pragma omp critical
-        {
-          for (const Astar::GridmapUpdate& u : updates)
-          {
-            if (g[u.getPos()] > u.getCost())
-            {
-              g[u.getPos()] = u.getCost();
-              open.push(std::move(u.getPriorityVec()));
-            }
-          }
-        }  // omp critical
-      }
-    }  // omp parallel
-    rough_cost_max_ = g[s_rough] + ec_[0] * (range_ + local_range_);
-  }
-  bool searchAvailablePos(Astar::Vec& s, const int xy_range, const int angle_range,
-                          const int cost_acceptable = 50, const int min_xy_range = 0)
+  template <class T>
+  bool searchAvailablePos(const T& cm, Astar::Vec& s, const int xy_range, const int angle_range,
+                          const int cost_acceptable = 50, const int min_xy_range = 0) const
   {
     ROS_DEBUG("%d, %d  (%d,%d,%d)", xy_range, angle_range, s[0], s[1], s[2]);
 
@@ -570,7 +476,7 @@ protected:
     {
       if (cost_acceptable != 100)
       {
-        return searchAvailablePos(s, xy_range, angle_range, 100);
+        return searchAvailablePos(cm, s, xy_range, angle_range, 100);
       }
       return false;
     }
@@ -591,11 +497,6 @@ protected:
       return true;
     }
 
-    prev_map_update_x_min_ = std::numeric_limits<int>::max();
-    prev_map_update_x_max_ = std::numeric_limits<int>::lowest();
-    prev_map_update_y_min_ = std::numeric_limits<int>::max();
-    prev_map_update_y_max_ = std::numeric_limits<int>::lowest();
-
     Astar::Vec s, e;
     grid_metric_converter::metric2Grid(
         map_info_, s[0], s[1], s[2],
@@ -612,63 +513,57 @@ protected:
       ROS_INFO(
           "New goal received (%d, %d, %d)",
           e[0], e[1], e[2]);
-    }
 
-    if (!cm_.validate(e, range_))
-    {
-      ROS_ERROR("Given goal is not on the map.");
-      return false;
+      clearHysteresis();
+      has_hysteresis_map_ = false;
     }
-    if (!cm_.validate(s, range_))
+    const DiscretePoseStatus start_pose_status = relocateDiscretePoseIfNeeded(s, tolerance_range_, tolerance_angle_);
+    const DiscretePoseStatus goal_pose_status = relocateDiscretePoseIfNeeded(e, tolerance_range_, tolerance_angle_);
+    switch (start_pose_status)
     {
-      ROS_ERROR("You are on the edge of the world.");
-      return false;
+      case DiscretePoseStatus::OUT_OF_MAP:
+        ROS_ERROR("You are on the edge of the world.");
+        return false;
+      case DiscretePoseStatus::IN_ROCK:
+        ROS_WARN("Oops! You are in Rock!");
+        ++cnt_stuck_;
+        return true;
+      default:
+        break;
     }
-
-    const auto ts = boost::chrono::high_resolution_clock::now();
-    reservable_priority_queue<Astar::PriorityVec> open;
-    open.reserve(map_info_.width * map_info_.height / 2);
-
-    cost_estim_cache_.clear(std::numeric_limits<float>::max());
-    if (cm_[e] == 100)
+    switch (goal_pose_status)
     {
-      if (!searchAvailablePos(e, tolerance_range_, tolerance_angle_))
-      {
+      case DiscretePoseStatus::OUT_OF_MAP:
+        ROS_ERROR("Given goal is not on the map.");
+        return false;
+      case DiscretePoseStatus::IN_ROCK:
         ROS_WARN("Oops! Goal is in Rock!");
         ++cnt_stuck_;
         return true;
-      }
-      ROS_INFO("Goal moved (%d, %d, %d)",
-               e[0], e[1], e[2]);
-      float x, y, yaw;
-      grid_metric_converter::grid2Metric(map_info_, e[0], e[1], e[2], x, y, yaw);
-      goal_.pose.orientation = tf2::toMsg(tf2::Quaternion(tf2::Vector3(0.0, 0.0, 1.0), yaw));
-      goal_.pose.position.x = x;
-      goal_.pose.position.y = y;
+      case DiscretePoseStatus::RELOCATED:
+        ROS_INFO("Goal moved (%d, %d, %d)", e[0], e[1], e[2]);
+        float x, y, yaw;
+        grid_metric_converter::grid2Metric(map_info_, e[0], e[1], e[2], x, y, yaw);
+        goal_.pose.orientation = tf2::toMsg(tf2::Quaternion(tf2::Vector3(0.0, 0.0, 1.0), yaw));
+        goal_.pose.position.x = x;
+        goal_.pose.position.y = y;
+        break;
+      default:
+        break;
     }
-    if (cm_[s] == 100)
-    {
-      if (!searchAvailablePos(s, tolerance_range_, tolerance_angle_))
-      {
-        ROS_WARN("Oops! You are in Rock!");
-        return true;
-      }
-    }
-
-    e[2] = 0;
-    cost_estim_cache_[e] = -ec_[0] * 0.5;  // Decrement to reduce calculation error
-    open.push(Astar::PriorityVec(cost_estim_cache_[e], cost_estim_cache_[e], e));
-    fillCostmap(open, cost_estim_cache_, s, e);
+    const auto ts = boost::chrono::high_resolution_clock::now();
+    cost_estim_cache_.create(s, e);
     const auto tnow = boost::chrono::high_resolution_clock::now();
-    ROS_DEBUG("Cost estimation cache generated (%0.4f sec.)",
-              boost::chrono::duration<float>(tnow - ts).count());
-    cost_estim_cache_[e] = 0;
-
-    if (goal_changed)
-    {
-      cm_hyst_.clear(100);
-      has_hysteresis_map_ = false;
-    }
+    const float dur = boost::chrono::duration<float>(tnow - ts).count();
+    ROS_DEBUG("Cost estimation cache generated (%0.4f sec.)", dur);
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "distance_map_init_dur",
+        dur,
+        "second"));
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "distance_map_update_dur",
+        0.0,
+        "second"));
 
     publishDebug();
 
@@ -676,6 +571,14 @@ protected:
 
     return true;
   }
+  void clearHysteresis()
+  {
+    for (const Astar::Vec& p : hyst_updated_cells_)
+    {
+      cm_hyst_[p] = 100;
+    }
+    hyst_updated_cells_.clear();
+  }
   void publishDebug()
   {
     if (pub_distance_map_.getNumSubscribers() > 0)
@@ -773,19 +676,92 @@ protected:
       pub_path_.publish(path);
     }
   }
+  void publishFinishPath()
+  {
+    nav_msgs::Path path;
+    path.header.frame_id = map_header_.frame_id;
+    path.header.stamp = ros::Time::now();
+    // Specify single pose to control only orientation
+    path.poses.resize(1);
+    path.poses[0].header = path.header;
+    if (force_goal_orientation_)
+      path.poses[0].pose = goal_raw_.pose;
+    else
+      path.poses[0].pose = goal_.pose;
 
-  void cbMapUpdate(const costmap_cspace_msgs::CSpace3DUpdate::ConstPtr& msg)
+    if (use_path_with_velocity_)
+    {
+      pub_path_velocity_.publish(
+          trajectory_tracker_msgs::toPathWithVelocity(
+              path, std::numeric_limits<double>::quiet_NaN()));
+    }
+    else
+    {
+      pub_path_.publish(path);
+    }
+  }
+
+  void cbMapUpdate(const costmap_cspace_msgs::CSpace3DUpdate::ConstPtr msg)
   {
     if (!has_map_)
       return;
     ROS_DEBUG("Map updated");
 
+    const auto ts_cm_init_start = boost::chrono::high_resolution_clock::now();
     const ros::Time now = ros::Time::now();
+
+    const int map_update_x_min = static_cast<int>(msg->x);
+    const int map_update_x_max = std::max(static_cast<int>(msg->x + msg->width) - 1, 0);
+    const int map_update_y_min = static_cast<int>(msg->y);
+    const int map_update_y_max = std::max(static_cast<int>(msg->y + msg->height) - 1, 0);
+
+    if (static_cast<size_t>(map_update_x_max) >= map_info_.width ||
+        static_cast<size_t>(map_update_y_max) >= map_info_.height ||
+        msg->angle > map_info_.angle)
+    {
+      ROS_WARN(
+          "Map update out of range (update range: %dx%dx%d, map range: %dx%dx%d)",
+          map_update_x_max, map_update_y_max, msg->angle,
+          map_info_.width, map_info_.height, map_info_.angle);
+      map_update_retained_ = msg;
+      return;
+    }
+    map_update_retained_ = nullptr;
+
     last_costmap_ = now;
 
-    cm_ = cm_base_;
-    cm_rough_ = cm_rough_base_;
-    cm_updates_.clear(-1);
+    cm_.copy_partially(
+        cm_base_,
+        Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
+        Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, static_cast<int>(map_info_.angle) - 1));
+    cm_rough_.copy_partially(
+        cm_rough_base_,
+        Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
+        Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, 0));
+    cm_updates_.clear_partially(
+        -1,
+        Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
+        Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, 0));
+
+    // Should search 1px around the region to
+    // update the costmap even if the edge of the local map is obstacle
+    const int search_range_x_min = std::max(
+        0,
+        std::min(prev_map_update_x_min_, map_update_x_min) - 1);
+    const int search_range_x_max = std::min(
+        static_cast<int>(map_info_.width - 1),
+        std::max(prev_map_update_x_max_, map_update_x_max) + 1);
+    const int search_range_y_min = std::max(
+        0,
+        std::min(prev_map_update_y_min_, map_update_y_min) - 1);
+    const int search_range_y_max = std::min(
+        static_cast<int>(map_info_.height - 1),
+        std::max(prev_map_update_y_max_, map_update_y_max) + 1);
+
+    prev_map_update_x_min_ = map_update_x_min;
+    prev_map_update_x_max_ = map_update_x_max;
+    prev_map_update_y_min_ = map_update_y_min;
+    prev_map_update_y_max_ = map_update_y_max;
 
     bool clear_hysteresis(false);
 
@@ -830,11 +806,18 @@ protected:
         }
       }
     }
+    const auto ts_cm_init_end = boost::chrono::high_resolution_clock::now();
+    const float ts_cm_init_dur = boost::chrono::duration<float>(ts_cm_init_end - ts_cm_init_start).count();
+    ROS_DEBUG("Costmaps updated (%.4f)", ts_cm_init_dur);
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "costmap_dur",
+        ts_cm_init_dur,
+        "second"));
 
     if (clear_hysteresis && has_hysteresis_map_)
     {
       ROS_INFO("The previous path collides to the obstacle. Clearing hysteresis map.");
-      cm_hyst_.clear(100);
+      clearHysteresis();
       has_hysteresis_map_ = false;
     }
 
@@ -879,107 +862,30 @@ protected:
       return;
     }
 
-    e[2] = 0;
-
     const auto ts = boost::chrono::high_resolution_clock::now();
-
-    const int map_update_x_min = static_cast<int>(msg->x);
-    const int map_update_x_max = static_cast<int>(msg->x + msg->width);
-    const int map_update_y_min = static_cast<int>(msg->y);
-    const int map_update_y_max = static_cast<int>(msg->y + msg->height);
-    const int search_range_x_min = std::max(0, std::min(prev_map_update_x_min_, map_update_x_min) - 1);
-    const int search_range_x_max = std::min(static_cast<int>(map_info_.width),
-                                            std::max(prev_map_update_x_max_, map_update_x_max) + 1);
-    const int search_range_y_min = std::max(0, std::min(prev_map_update_y_min_, map_update_y_min) - 1);
-    const int search_range_y_max = std::min(static_cast<int>(map_info_.height),
-                                            std::max(prev_map_update_y_max_, map_update_y_max) + 1);
-    Astar::Vec p, p_cost_min;
-    p[2] = 0;
-    float cost_min = std::numeric_limits<float>::max();
-    for (p[1] = search_range_y_min; p[1] < search_range_y_max; p[1]++)
-    {
-      for (p[0] = search_range_x_min; p[0] < search_range_x_max; p[0]++)
-      {
-        if (cost_min > cost_estim_cache_[p])
-        {
-          p_cost_min = p;
-          cost_min = cost_estim_cache_[p];
-        }
-      }
-    }
-    prev_map_update_x_min_ = map_update_x_min;
-    prev_map_update_x_max_ = map_update_x_max;
-    prev_map_update_y_min_ = map_update_y_min;
-    prev_map_update_y_max_ = map_update_y_max;
-
-    reservable_priority_queue<Astar::PriorityVec> open;
-    reservable_priority_queue<Astar::PriorityVec> erase;
-    open.reserve(map_info_.width * map_info_.height / 2);
-    erase.reserve(map_info_.width * map_info_.height / 2);
-
-    if (cost_min != std::numeric_limits<float>::max())
-      erase.emplace(cost_min, cost_min, p_cost_min);
-    while (true)
-    {
-      if (erase.size() < 1)
-        break;
-      const Astar::PriorityVec center(erase.top());
-      const Astar::Vec p = center.v_;
-      erase.pop();
-
-      if (cost_estim_cache_[p] == std::numeric_limits<float>::max())
-        continue;
-      cost_estim_cache_[p] = std::numeric_limits<float>::max();
-
-      Astar::Vec d;
-      d[2] = 0;
-      for (d[0] = -1; d[0] <= 1; d[0]++)
-      {
-        for (d[1] = -1; d[1] <= 1; d[1]++)
-        {
-          if (!((d[0] == 0) ^ (d[1] == 0)))
-            continue;
-          Astar::Vec next = p + d;
-          next[2] = 0;
-          if ((unsigned int)next[0] >= (unsigned int)map_info_.width ||
-              (unsigned int)next[1] >= (unsigned int)map_info_.height)
-            continue;
-          const float gn = cost_estim_cache_[next];
-          if (gn == std::numeric_limits<float>::max())
-            continue;
-          if (gn < cost_min)
-          {
-            open.emplace(gn, gn, next);
-            continue;
-          }
-          erase.emplace(gn, gn, next);
-        }
-      }
-    }
-    if (open.size() == 0)
-    {
-      open.emplace(-ec_[0] * 0.5, -ec_[0] * 0.5, e);
-    }
-    {
-      Astar::Vec p;
-      p[2] = 0;
-      for (p[0] = 0; p[0] < static_cast<int>(map_info_.width); p[0]++)
-      {
-        for (p[1] = 0; p[1] < static_cast<int>(map_info_.height); p[1]++)
-        {
-          const auto gp = cost_estim_cache_[p];
-          if (gp > rough_cost_max_)
-          {
-            open.emplace(gp, gp, p);
-          }
-        }
-      }
-    }
-
-    fillCostmap(open, cost_estim_cache_, s, e);
+    cost_estim_cache_.update(
+        s, e,
+        DistanceMap::Rect(
+            Astar::Vec(search_range_x_min, search_range_y_min, 0),
+            Astar::Vec(search_range_x_max, search_range_y_max, 0)));
     const auto tnow = boost::chrono::high_resolution_clock::now();
-    ROS_DEBUG("Cost estimation cache updated (%0.4f sec.)",
-              boost::chrono::duration<float>(tnow - ts).count());
+    const DistanceMap::DebugData dm_debug = cost_estim_cache_.getDebugData();
+    if (dm_debug.has_negative_cost)
+    {
+      ROS_WARN("Negative cost value is detected. Limited to zero.");
+    }
+    ROS_DEBUG("Cost estimation cache search queue initial size: %lu, capacity: %lu",
+              dm_debug.search_queue_size, dm_debug.search_queue_cap);
+    const float dur = boost::chrono::duration<float>(tnow - ts).count();
+    ROS_DEBUG("Cost estimation cache updated (%0.4f sec.)", dur);
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "distance_map_update_dur",
+        dur,
+        "second"));
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "distance_map_init_dur",
+        0.0,
+        "second"));
     publishDebug();
   }
   void cbMap(const costmap_cspace_msgs::CSpace3D::ConstPtr& msg)
@@ -1011,7 +917,6 @@ protected:
       hist_ignore_range_ = std::lround(hist_ignore_range_f_ / map_info_.linear_resolution);
       hist_ignore_range_max_ = std::lround(hist_ignore_range_max_f_ / map_info_.linear_resolution);
       local_range_ = std::lround(local_range_f_ / map_info_.linear_resolution);
-      longcut_range_ = std::lround(longcut_range_f_ / map_info_.linear_resolution);
       esc_range_ = std::lround(esc_range_f_ / map_info_.linear_resolution);
       esc_angle_ = map_info_.angle / 8;
       tolerance_range_ = std::lround(tolerance_range_f_ / map_info_.linear_resolution);
@@ -1025,7 +930,7 @@ protected:
               map_info_,
               ec_,
               local_range_,
-              cost_estim_cache_, cm_, cm_hyst_, cm_rough_,
+              cost_estim_cache_.gridmap(), cm_, cm_hyst_, cm_rough_,
               cc_, range_));
 
       ROS_DEBUG("Search model updated");
@@ -1039,18 +944,31 @@ protected:
 
     const int size[3] =
         {
-          static_cast<int>(map_info_.width),
-          static_cast<int>(map_info_.height),
-          static_cast<int>(map_info_.angle)
+            static_cast<int>(map_info_.width),
+            static_cast<int>(map_info_.height),
+            static_cast<int>(map_info_.angle),
         };
-    as_.reset(Astar::Vec(size[0], size[1], size[2]));
-    cm_.reset(Astar::Vec(size[0], size[1], size[2]));
-    cm_hyst_.reset(Astar::Vec(size[0], size[1], size[2]));
 
-    cost_estim_cache_.reset(Astar::Vec(size[0], size[1], 1));
-    cm_rough_.reset(Astar::Vec(size[0], size[1], 1));
-    cm_updates_.reset(Astar::Vec(size[0], size[1], 1));
-    bbf_costmap_.reset(Astar::Vec(size[0], size[1], 1));
+    const Astar::Vec size3d(size[0], size[1], size[2]);
+    const Astar::Vec size2d(size[0], size[1], 1);
+
+    as_.reset(size3d);
+    cm_.reset(size3d);
+    cm_hyst_.reset(size3d);
+
+    const DistanceMap::Params dmp =
+        {
+            .euclid_cost = ec_,
+            .range = range_,
+            .local_range = local_range_,
+            .longcut_range = static_cast<int>(std::lround(longcut_range_f_ / map_info_.linear_resolution)),
+            .size = size2d,
+            .resolution = map_info_.linear_resolution,
+        };
+    cost_estim_cache_.init(model_, dmp);
+    cm_rough_.reset(size2d);
+    cm_updates_.reset(size2d);
+    bbf_costmap_.reset(size2d);
 
     Astar::Vec p;
     for (p[0] = 0; p[0] < static_cast<int>(map_info_.width); p[0]++)
@@ -1075,15 +993,30 @@ protected:
     ROS_DEBUG("Map copied");
 
     cm_hyst_.clear(100);
+    hyst_updated_cells_.clear();
     has_hysteresis_map_ = false;
 
+    cm_updates_.clear(0);
+
     has_map_ = true;
 
     cm_rough_base_ = cm_rough_;
     cm_base_ = cm_;
     bbf_costmap_.clear();
 
+    prev_map_update_x_min_ = map_info_.width;
+    prev_map_update_x_max_ = 0;
+    prev_map_update_y_min_ = map_info_.height;
+    prev_map_update_y_max_ = 0;
+
     updateGoal();
+
+    if (map_update_retained_ && map_update_retained_->header.stamp >= msg->header.stamp)
+    {
+      ROS_INFO("Applying retained map update");
+      cbMapUpdate(map_update_retained_);
+    }
+    map_update_retained_ = nullptr;
   }
   void cbAction()
   {
@@ -1143,6 +1076,7 @@ public:
     : nh_()
     , pnh_("~")
     , tfl_(tfbuf_)
+    , cost_estim_cache_(cm_rough_, bbf_costmap_)
     , jump_(tfbuf_)
   {
     neonavigation_common::compat::checkCompatMode();
@@ -1158,6 +1092,7 @@ public:
     pub_start_ = pnh_.advertise<geometry_msgs::PoseStamped>("path_start", 1, true);
     pub_end_ = pnh_.advertise<geometry_msgs::PoseStamped>("path_end", 1, true);
     pub_status_ = pnh_.advertise<planner_cspace_msgs::PlannerStatus>("status", 1, true);
+    pub_metrics_ = pnh_.advertise<neonavigation_metrics_msgs::Metrics>("metrics", 1, false);
     srs_forget_ = neonavigation_common::compat::advertiseService(
         nh_, "forget_planning_cost",
         pnh_, "forget", &Planner3dNode::cbForget, this);
@@ -1194,6 +1129,7 @@ public:
 
     pnh_.param("freq", freq_, 4.0f);
     pnh_.param("freq_min", freq_min_, 2.0f);
+    pnh_.param("search_timeout_abort", search_timeout_abort_, 30.0f);
     pnh_.param("search_range", search_range_, 0.4f);
     pnh_.param("antialias_start", antialias_start_, false);
 
@@ -1290,7 +1226,9 @@ public:
     int num_task;
     pnh_.param("num_search_task", num_task, num_threads * 16);
     as_.setSearchTaskNum(num_task);
-    pnh_.param("num_cost_estim_task", num_cost_estim_task_, num_threads * 16);
+    int num_cost_estim_task;
+    pnh_.param("num_cost_estim_task", num_cost_estim_task, num_threads * 16);
+    cost_estim_cache_.setParams(cc_, num_cost_estim_task);
 
     pnh_.param("retain_last_error_status", retain_last_error_status_, true);
     status_.status = planner_cspace_msgs::PlannerStatus::DONE;
@@ -1387,6 +1325,7 @@ public:
       waitUntil(next_replan_time, previous_path);
 
       const ros::Time now = ros::Time::now();
+      next_replan_time = now;
 
       if (has_map_ && !goal_updated_ && has_goal_)
       {
@@ -1395,7 +1334,12 @@ public:
       bool has_costmap(false);
       if (costmap_watchdog_ > ros::Duration(0))
       {
-        if (last_costmap_ + costmap_watchdog_ < now)
+        const ros::Duration costmap_delay = now - last_costmap_;
+        metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+            "costmap_delay",
+            costmap_delay.toSec(),
+            "second"));
+        if (costmap_delay > costmap_watchdog_)
         {
           ROS_WARN_THROTTLE(1.0,
                             "Navigation is stopping since the costmap is too old (costmap: %0.3f)",
@@ -1411,6 +1355,10 @@ public:
       }
       else
       {
+        metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+            "costmap_delay",
+            -1.0,
+            "second"));
         has_costmap = true;
       }
 
@@ -1457,6 +1405,10 @@ public:
             if (act_tolerant_->isActive())
               act_tolerant_->setSucceeded(planner_cspace_msgs::MoveWithToleranceResult(), "Goal reached.");
           }
+          else
+          {
+            publishFinishPath();
+          }
         }
         else
         {
@@ -1484,6 +1436,10 @@ public:
 
             continue;
           }
+          else if (cnt_stuck_ > 0)
+          {
+            status_.error = planner_cspace_msgs::PlannerStatus::PATH_NOT_FOUND;
+          }
           else
           {
             status_.error = planner_cspace_msgs::PlannerStatus::GOING_WELL;
@@ -1520,9 +1476,26 @@ public:
         publishEmptyPath();
         previous_path.poses.clear();
       }
+      status_.header.stamp = now;
       pub_status_.publish(status_);
       diag_updater_.force_update();
 
+      metrics_.header.stamp = now;
+      metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+          "stuck_cnt",
+          cnt_stuck_,
+          "count"));
+      metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+          "error",
+          status_.error,
+          "enum"));
+      metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+          "status",
+          status_.status,
+          "enum"));
+      pub_metrics_.publish(metrics_);
+      metrics_.data.clear();
+
       if (is_path_switchback)
       {
         next_replan_time += ros::Duration(sw_wait_);
@@ -1561,14 +1534,8 @@ protected:
     std::vector<Astar::VecWithCost> starts;
     if (antialias_start_)
     {
-      const int x_cand[] =
-          {
-            0, ((sf[0] - s[0]) < 0.5 ? -1 : 1)
-          };
-      const int y_cand[] =
-          {
-            0, ((sf[1] - s[1]) < 0.5 ? -1 : 1)
-          };
+      const int x_cand[] = {0, ((sf[0] - s[0]) < 0.5 ? -1 : 1)};
+      const int y_cand[] = {0, ((sf[1] - s[1]) < 0.5 ? -1 : 1)};
       for (const int x : x_cand)
       {
         for (const int y : y_cand)
@@ -1626,20 +1593,25 @@ protected:
           escaping_ = false;
           updateGoal();
           ROS_INFO("Escaped");
+          return true;
         }
         else
         {
-          path.poses.resize(1);
-          path.poses[0].header = path.header;
-          if (force_goal_orientation_)
-            path.poses[0].pose = goal_raw_.pose;
+          if (act_tolerant_->isActive() && goal_tolerant_->continuous_movement_mode)
+          {
+            ROS_INFO("Robot reached near the goal.");
+            act_tolerant_->setSucceeded(planner_cspace_msgs::MoveWithToleranceResult(),
+                                        "Goal reached (Continuous movement mode).");
+            goal_tolerant_ = nullptr;
+          }
           else
-            path.poses[0].pose = ge;
-
-          status_.status = planner_cspace_msgs::PlannerStatus::FINISHING;
-          ROS_INFO("Path plan finishing");
+          {
+            status_.status = planner_cspace_msgs::PlannerStatus::FINISHING;
+            publishFinishPath();
+            ROS_INFO("Path plan finishing");
+            return true;
+          }
         }
-        return true;
       }
     }
 
@@ -1654,7 +1626,7 @@ protected:
 
     if (starts.size() == 0)
     {
-      if (!searchAvailablePos(s, tolerance_range_, tolerance_angle_))
+      if (!searchAvailablePos(cm_, s, tolerance_range_, tolerance_angle_))
       {
         ROS_WARN("Oops! You are in Rock!");
         status_.error = planner_cspace_msgs::PlannerStatus::IN_ROCK;
@@ -1665,14 +1637,16 @@ protected:
     }
     const Astar::Vec s_rough(s[0], s[1], 0);
 
-    if (cost_estim_cache_[s_rough] == std::numeric_limits<float>::max())
+    // If goal gets occupied, cost_estim_cache_ is not updated to reduce
+    // computational cost for clearing huge map. In this case, cm_[e] is 100.
+    if (cost_estim_cache_[s_rough] == std::numeric_limits<float>::max() || cm_[e] >= 100)
     {
       status_.error = planner_cspace_msgs::PlannerStatus::PATH_NOT_FOUND;
       ROS_WARN("Goal unreachable.");
       if (!escaping_ && temporary_escape_)
       {
         e = s;
-        if (searchAvailablePos(e, esc_range_, esc_angle_, 50, esc_range_ / 2))
+        if (searchAvailablePos(cm_, e, esc_range_, esc_angle_, 50, esc_range_ / 2))
         {
           escaping_ = true;
           ROS_INFO("Temporary goal (%d, %d, %d)",
@@ -1702,12 +1676,34 @@ protected:
     // ROS_INFO("Planning from (%d, %d, %d) to (%d, %d, %d)",
     //   s[0], s[1], s[2], e[0], e[1], e[2]);
 
+    const auto cb_progress = [this, ts, s, e](const std::list<Astar::Vec>& path_grid, const SearchStats& stats) -> bool
+    {
+      const auto tnow = boost::chrono::high_resolution_clock::now();
+      const auto tdiff = boost::chrono::duration<float>(tnow - ts).count();
+      publishEmptyPath();
+      if (tdiff > search_timeout_abort_)
+      {
+        ROS_ERROR(
+            "Search aborted due to timeout. "
+            "search_timeout_abort may be too small or planner_3d may have a bug: "
+            "s=(%d, %d, %d), g=(%d, %d, %d), tdiff=%0.4f, "
+            "num_loop=%lu, "
+            "num_search_queue=%lu, "
+            "num_prev_updates=%lu, "
+            "num_total_updates=%lu, ",
+            s[0], s[1], s[2], e[0], e[1], e[2], tdiff,
+            stats.num_loop, stats.num_search_queue, stats.num_prev_updates, stats.num_total_updates);
+        return false;
+      }
+      ROS_WARN("Search timed out (%0.4f sec.)", tdiff);
+      return true;
+    };
+
     model_->enableHysteresis(hyst && has_hysteresis_map_);
     std::list<Astar::Vec> path_grid;
     if (!as_.search(starts, e, path_grid,
                     model_,
-                    std::bind(&Planner3dNode::cbProgress,
-                              this, std::placeholders::_1),
+                    cb_progress,
                     range_limit,
                     1.0f / freq_min_,
                     find_best_))
@@ -1718,8 +1714,12 @@ protected:
         return false;
     }
     const auto tnow = boost::chrono::high_resolution_clock::now();
-    ROS_DEBUG("Path found (%0.4f sec.)",
-              boost::chrono::duration<float>(tnow - ts).count());
+    const float dur = boost::chrono::duration<float>(tnow - ts).count();
+    ROS_DEBUG("Path found (%0.4f sec.)", dur);
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "path_search_dur",
+        dur,
+        "second"));
 
     geometry_msgs::PoseArray poses;
     poses.header = path.header;
@@ -1742,6 +1742,7 @@ protected:
 
     if (hyst)
     {
+      const auto ts = boost::chrono::high_resolution_clock::now();
       std::unordered_map<Astar::Vec, bool, Astar::Vec> path_points;
       const float max_dist = cc_.hysteresis_max_dist_ / map_info_.linear_resolution;
       const float expand_dist = cc_.hysteresis_expand_ / map_info_.linear_resolution;
@@ -1763,8 +1764,7 @@ protected:
         }
       }
 
-      cm_hyst_.clear(100);
-      const auto ts = boost::chrono::high_resolution_clock::now();
+      clearHysteresis();
       for (auto& ps : path_points)
       {
         const Astar::Vec& p = ps.first;
@@ -1792,23 +1792,22 @@ protected:
         }
         d_min = std::max(expand_dist, std::min(expand_dist + max_dist, d_min));
         cm_hyst_[p] = std::lround((d_min - expand_dist) * 100.0 / max_dist);
+        hyst_updated_cells_.push_back(p);
       }
       has_hysteresis_map_ = true;
       const auto tnow = boost::chrono::high_resolution_clock::now();
-      ROS_DEBUG("Hysteresis map generated (%0.4f sec.)",
-                boost::chrono::duration<float>(tnow - ts).count());
+      const float dur = boost::chrono::duration<float>(tnow - ts).count();
+      ROS_DEBUG("Hysteresis map generated (%0.4f sec.)", dur);
+      metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+          "hyst_map_dur",
+          dur,
+          "second"));
       publishDebug();
     }
 
     return true;
   }
 
-  bool cbProgress(const std::list<Astar::Vec>& path_grid)
-  {
-    publishEmptyPath();
-    ROS_WARN("Search timed out");
-    return true;
-  }
   int getSwitchIndex(const nav_msgs::Path& path) const
   {
     geometry_msgs::Pose p_prev;
